import base64
import asyncio
import traceback
import urllib.parse
import typing
import json
import logging
from dataclasses import dataclass

from mangum.lifespan import Lifespan
from mangum.utils import get_logger, make_response, get_server_and_client
from mangum.types import ASGIApp


@dataclass
class Mangum:

    app: ASGIApp
    debug: bool = False
    enable_lifespan: bool = True
    log_level: str = "info"

    def __post_init__(self) -> None:
        self.logger = get_logger(log_level=self.log_level)
        if self.enable_lifespan:
            loop = asyncio.get_event_loop()
            self.lifespan = Lifespan(self.app, logger=self.logger)
            loop.create_task(self.lifespan.run())
            loop.run_until_complete(self.lifespan.wait_startup())

    def __call__(
        self, event: typing.Dict[str, typing.Any], context: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Any]:
        try:
            response = self.handler(event, context)
        except Exception as exc:
            if self.debug:
                content = traceback.format_exc()
                return make_response(content, status_code=500)
            raise exc
        else:
            return response

    def handler(
        self, event: typing.Dict[str, typing.Any], context: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Any]:
        if "httpMethod" in event:
            from mangum.protocols.http import handle_http

            response = handle_http(self.app, event, context)
        else:
            try:
                from mangum.protocols.websockets import handle_ws
            except ImportError as exc:  # pragma: nocover
                raise exc

            response = handle_ws(self.app, event, context)

        if self.enable_lifespan:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self.lifespan.wait_shutdown())
        return response



import typing
import json
from dataclasses import dataclass

from mangum.utils import make_response, get_server_and_client
from mangum.connections import ConnectionTable
from mangum.protocols.asgi import ASGICycle, ASGICycleState
from mangum.types import ASGIMessage, ASGIApp
from mangum.exceptions import ASGIWebSocketCycleException


@dataclass
class ASGIWebSocketCycle(ASGICycle):

    endpoint_url: str = None
    connection_id: str = None
    connection_table: ConnectionTable = None

    async def asgi_send(self, message: ASGIMessage) -> None:
        if self.state is ASGICycleState.REQUEST:
            if message["type"] in ("websocket.accept", "websocket.close"):
                self.state = ASGICycleState.RESPONSE
            else:
                raise RuntimeError(
                    f"Expected 'websocket.accept' or 'websocket.close', received: {message['type']}"
                )
        else:
            data = message.get("text")
            if message["type"] == "websocket.send":
                group = message.get("group", None)
                self.send_data(data=data, group=group)

    def send_data(self, *, data: str, group: str = None) -> None:  # pragma: no cover
        """
        Send a data message to a client or group of clients using the connection table.
        """
        item = self.connection_table.get_item(self.connection_id)
        if group:
            # Retrieve the existing groups for the current connection, or create a new
            # groups entry if one does not exist.
            groups = item.get("groups", [])
            if group not in groups:
                # Ensure the group specified in the message is included.
                groups.append(group)
                result = self.connection_table.update_item(
                    self.connection_id, groups=groups
                )
                status_code = result.get("ResponseMetadata", {}).get("HTTPStatusCode")
                if status_code != 200:
                    raise ASGIWebSocketCycleException("Error updating groups")

            # Retrieve all items associated with the current group.
            items = self.connection_table.get_group_items(group)
            if items is None:
                raise ASGIWebSocketCycleException("No connections found")
        else:
            # Single send, add the current item to a list to be iterated by the
            # connection table.
            items = [item]
        self.connection_table.send_data(items, data=data)


def handle_ws(
    app: ASGIApp,
    event: typing.Dict[str, typing.Any],
    context: typing.Dict[str, typing.Any],
) -> typing.Dict[str, typing.Any]:
    request_context = event["requestContext"]
    connection_id = request_context.get("connectionId")
    domain_name = request_context.get("domainName")
    stage = request_context.get("stage")
    event_type = request_context["eventType"]
    endpoint_url = f"https://{domain_name}/{stage}"

    if event_type == "CONNECT":
        # The initial connect event. Parse and store the scope for the connection
        # in DynamoDB to be retrieved in subsequent message events for this request.
        server, client = get_server_and_client(event)
        headers = event["headers"]
        root_path = event["requestContext"]["stage"]
        scope = {
            "type": "websocket",
            "path": "/",
            "headers": headers,
            "raw_path": None,
            "root_path": root_path,
            "scheme": event["headers"].get("X-Forwarded-Proto", "wss"),
            "query_string": "",
            "server": server,
            "client": client,
        }
        connection_table = ConnectionTable()
        status_code = connection_table.update_item(
            connection_id, scope=json.dumps(scope)
        )
        if status_code != 200:  # pragma: no cover
            # TODO: Improve error handling
            return make_response("Error", status_code=500)
        return make_response("OK", status_code=200)

    elif event_type == "MESSAGE":
        event_body = json.loads(event["body"])
        event_data = event_body["data"] or ""

        connection_table = ConnectionTable()
        item = connection_table.get_item(connection_id)
        if not item:  # pragma: no cover
            # TODO: Improve error handling
            return make_response("Error", status_code=500)

        # Retrieve and deserialize the scope entry created in the connect event for
        # the current connection.
        scope = json.loads(item["scope"])

        # Ensure the scope definitions comply with the ASGI spec.
        query_string = scope["query_string"]
        headers = scope["headers"]
        headers = [[k.encode(), v.encode()] for k, v in headers.items()]
        query_string = query_string.encode()
        scope.update({"headers": headers, "query_string": query_string})

        asgi_cycle = ASGIWebSocketCycle(
            scope,
            endpoint_url=endpoint_url,
            connection_id=connection_id,
            connection_table=connection_table,
        )
        asgi_cycle.put_message({"type": "websocket.connect"})
        asgi_cycle.put_message(
            {
                "type": "websocket.receive",
                "path": "/",
                "bytes": None,
                "text": event_data,
            }
        )
        try:
            asgi_cycle(app)
        except ASGIWebSocketCycleException:  # pragma: no cover
            # TODO: Improve error handling
            return make_response("Error", status_code=500)
        return make_response("OK", status_code=200)

    elif event_type == "DISCONNECT":
        connection_table = ConnectionTable()
        status_code = connection_table.delete_item(connection_id)
        if status_code != 200:  # pragma: no cover
            # TODO: Improve error handling
            return make_response("WebSocket disconnect error.", status_code=500)
        return make_response("OK", status_code=200)



import enum
import asyncio
import base64
import typing
from dataclasses import dataclass, field

from mangum.types import ASGIScope, ASGIMessage, ASGIApp


class ASGICycleState(enum.Enum):
    REQUEST = enum.auto()
    RESPONSE = enum.auto()


@dataclass
class ASGICycle:

    scope: ASGIScope
    state: ASGICycleState = ASGICycleState.REQUEST
    binary: bool = False
    response: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.loop = asyncio.get_event_loop()
        self.app_queue = asyncio.Queue(loop=self.loop)

    def __call__(self, app: ASGIApp) -> typing.Dict[str, typing.Any]:
        asgi_instance = app(self.scope, self.asgi_receive, self.asgi_send)
        asgi_task = self.loop.create_task(asgi_instance)
        self.loop.run_until_complete(asgi_task)
        return self.response

    async def asgi_receive(self) -> ASGIMessage:
        message = await self.app_queue.get()
        return message

    def put_message(self, message: ASGIMessage) -> None:
        self.app_queue.put_nowait(message)



import urllib.parse
import base64
import typing
from dataclasses import dataclass

from mangum.protocols.asgi import ASGICycle, ASGICycleState
from mangum.types import ASGIMessage, ASGIApp
from mangum.exceptions import ASGIWebSocketCycleException
from mangum.utils import get_server_and_client


@dataclass
class ASGIHTTPCycle(ASGICycle):

    body: bytes = b""

    async def asgi_send(self, message: ASGIMessage) -> None:
        if self.state is ASGICycleState.REQUEST:
            if message["type"] != "http.response.start":
                raise RuntimeError(
                    f"Expected 'http.response.start', received: {message['type']}"
                )

            status_code = message["status"]
            headers = {k: v for k, v in message.get("headers", [])}
            self.response["statusCode"] = status_code
            self.response["isBase64Encoded"] = self.binary
            self.response["headers"] = {
                k.decode(): v.decode() for k, v in headers.items()
            }
            self.state = ASGICycleState.RESPONSE

        elif self.state is ASGICycleState.RESPONSE:
            if message["type"] != "http.response.body":
                raise RuntimeError(
                    f"Expected 'http.response.body', received: {message['type']}"
                )

            body = message.get("body", b"")
            more_body = message.get("more_body", False)

            # The body must be completely read before returning the response.
            self.body += body

            if not more_body:
                body = self.body
                if self.binary:
                    body = base64.b64encode(body)
                self.response["body"] = body.decode()
                self.put_message({"type": "http.disconnect"})


def handle_http(
    app: ASGIApp,
    event: typing.Dict[str, typing.Any],
    context: typing.Dict[str, typing.Any],
) -> typing.Dict[str, typing.Any]:
    server, client = get_server_and_client(event)
    headers = [[k.lower().encode(), v.encode()] for k, v in event["headers"].items()]
    query_string_params = event["queryStringParameters"]
    query_string = (
        urllib.parse.urlencode(query_string_params).encode()
        if query_string_params
        else b""
    )
    scope = {
        "type": "http",
        "http_version": "1.1",
        "method": event["httpMethod"],
        "headers": headers,
        "path": urllib.parse.unquote(event["path"]),
        "raw_path": None,
        "root_path": "",
        "scheme": event["headers"].get("X-Forwarded-Proto", "https"),
        "query_string": query_string,
        "server": server,
        "client": client,
    }
    binary = event.get("isBase64Encoded", False)
    body = event["body"] or b""
    if binary:
        body = base64.b64decode(body)
    elif not isinstance(body, bytes):
        body = body.encode()
    asgi_cycle = ASGIHTTPCycle(scope, binary=binary)
    asgi_cycle.put_message({"type": "http.request", "body": body, "more_body": False})
    response = asgi_cycle(app)
    return response
